import os from 'os';
import { parse, Kind, print, isExecutableDefinitionNode } from 'graphql';
import { uniqueCode } from 'webql-tools-webpack-loader-runtime/es5';
import { optimizeDocumentNode, removeDescriptions, removeEmptyNodes } from 'webql-tools-optimize/es5';

/**
 * Strip insignificant whitespace
 * Note that this could do a lot more, such as reorder fields etc.
 */
function normalize(str) {
    return str.replace(/[\s,]+/g, ' ').trim();
}
// A map docString -> graphql document
var docCache = {};
// A map fragmentName -> [normalized source]
var fragmentSourceMap = {};
function cacheKeyFromFragment(fragment) {
    return normalize(print(fragment));
}
/**
 * Take a unstripped parsed document (query/mutation or even fragment), and
 * check all fragment definitions, checking for name->source uniqueness.
 * We also want to make sure only unique fragments exist in the document.
 */
function processFragments(ast) {
    var astFragmentMap = {};
    var definitions = [];
    for (var i = 0; i < ast.definitions.length; i++) {
        var fragmentDefinition = ast.definitions[i];
        if (fragmentDefinition.kind === Kind.FRAGMENT_DEFINITION) {
            var fragmentName = fragmentDefinition.name.value;
            var sourceKey = cacheKeyFromFragment(fragmentDefinition);
            // We know something about this fragment
            if (fragmentSourceMap.hasOwnProperty(fragmentName) && !fragmentSourceMap[fragmentName][sourceKey]) {
                fragmentSourceMap[fragmentName][sourceKey] = true;
            }
            else if (!fragmentSourceMap.hasOwnProperty(fragmentName)) {
                fragmentSourceMap[fragmentName] = {};
                fragmentSourceMap[fragmentName][sourceKey] = true;
            }
            if (!astFragmentMap[sourceKey]) {
                astFragmentMap[sourceKey] = true;
                definitions.push(fragmentDefinition);
            }
        }
        else {
            definitions.push(fragmentDefinition);
        }
    }
    ast.definitions = definitions;
    return ast;
}
function parseDocument(doc) {
    var cacheKey = normalize(doc);
    if (docCache[cacheKey]) {
        return docCache[cacheKey];
    }
    var parsed = parse(doc, {
        noLocation: true,
    });
    if (!parsed || parsed.kind !== 'Document') {
        throw new Error('Not a valid GraphQL document.');
    }
    // check that all "new" fragments inside the documents are consistent with
    // existing fragments of the same name
    docCache[cacheKey] = processFragments(parsed);
    return parsed;
}

function isSDL(doc) {
    return !doc.definitions.some(function (def) { return isExecutableDefinitionNode(def); });
}
function expandImports(source, options) {
    var lines = source.split(/\r\n|\r|\n/);
    var outputCode = options.importHelpers
        ? "\n    var useUnique = require('webql-tools-webpack-loader-runtime/es5').useUnique;\n    var unique = useUnique();\n  "
        : "\n    " + uniqueCode + "\n  ";
    lines.some(function (line) {
        if (line[0] === '#' && line.slice(1).split(' ')[0] === 'import') {
            var importFile = line.slice(1).split(' ')[1];
            var parseDocument_1 = "require(" + importFile + ")";
            var appendDef = "doc.definitions = doc.definitions.concat(unique(" + parseDocument_1 + ".definitions));";
            outputCode += appendDef + os.EOL;
        }
        return line.length !== 0 && line[0] !== '#';
    });
    return outputCode;
}
function graphqlLoader(source) {
    this.cacheable();
    var options = this.query || {};
    var doc = parseDocument(source);
    var optimizers = [];
    if (options.noDescription) {
        optimizers.push(removeDescriptions);
    }
    if (options.noEmptyNodes) {
        optimizers.push(removeEmptyNodes);
    }
    if (optimizers.length > 0 && isSDL(doc)) {
        doc = optimizeDocumentNode(doc, optimizers);
    }
    var headerCode = "\n    var doc = " + JSON.stringify(doc) + ";\n  ";
    var outputCode = '';
    // Allow multiple query/mutation definitions in a file. This parses out dependencies
    // at compile time, and then uses those at load time to create minimal query documents
    // We cannot do the latter at compile time due to how the #import code works.
    var operationCount = doc.definitions.reduce(function (accum, op) {
        if (op.kind === Kind.OPERATION_DEFINITION) {
            return accum + 1;
        }
        return accum;
    }, 0);
    function exportDefaultStatement(identifier) {
        if (options.esModule) {
            return "export default " + identifier;
        }
        return "module.exports = " + identifier;
    }
    if (operationCount > 1) {
        throw new Error('GraphQL Webpack Loader allows only for one GraphQL Operation per file');
    }
    outputCode += "\n    " + exportDefaultStatement('doc') + "\n  ";
    var importOutputCode = expandImports(source, options);
    var allCode = [headerCode, importOutputCode, outputCode, ''].join(os.EOL);
    return allCode;
}

export default graphqlLoader;
//# sourceMappingURL=index.esm.js.map
