import os from 'os';
import { parse, Kind, print, isExecutableDefinitionNode } from 'graphql';
import { uniqueCode } from 'webql-tools-webpack-loader-runtime';
import { optimizeDocumentNode, removeDescriptions, removeEmptyNodes } from 'webql-tools-optimize';

/**
 * Strip insignificant whitespace
 * Note that this could do a lot more, such as reorder fields etc.
 */
function normalize(str) {
    return str.replace(/[\s,]+/g, ' ').trim();
}
// A map docString -> graphql document
const docCache = {};
// A map fragmentName -> [normalized source]
const fragmentSourceMap = {};
function cacheKeyFromFragment(fragment) {
    return normalize(print(fragment));
}
/**
 * Take a unstripped parsed document (query/mutation or even fragment), and
 * check all fragment definitions, checking for name->source uniqueness.
 * We also want to make sure only unique fragments exist in the document.
 */
function processFragments(ast) {
    const astFragmentMap = {};
    const definitions = [];
    for (let i = 0; i < ast.definitions.length; i++) {
        const fragmentDefinition = ast.definitions[i];
        if (fragmentDefinition.kind === Kind.FRAGMENT_DEFINITION) {
            const fragmentName = fragmentDefinition.name.value;
            const sourceKey = cacheKeyFromFragment(fragmentDefinition);
            // We know something about this fragment
            if (fragmentSourceMap.hasOwnProperty(fragmentName) && !fragmentSourceMap[fragmentName][sourceKey]) {
                fragmentSourceMap[fragmentName][sourceKey] = true;
            }
            else if (!fragmentSourceMap.hasOwnProperty(fragmentName)) {
                fragmentSourceMap[fragmentName] = {};
                fragmentSourceMap[fragmentName][sourceKey] = true;
            }
            if (!astFragmentMap[sourceKey]) {
                astFragmentMap[sourceKey] = true;
                definitions.push(fragmentDefinition);
            }
        }
        else {
            definitions.push(fragmentDefinition);
        }
    }
    ast.definitions = definitions;
    return ast;
}
function parseDocument(doc) {
    const cacheKey = normalize(doc);
    if (docCache[cacheKey]) {
        return docCache[cacheKey];
    }
    const parsed = parse(doc, {
        noLocation: true,
    });
    if (!parsed || parsed.kind !== 'Document') {
        throw new Error('Not a valid GraphQL document.');
    }
    // check that all "new" fragments inside the documents are consistent with
    // existing fragments of the same name
    docCache[cacheKey] = processFragments(parsed);
    return parsed;
}

function isSDL(doc) {
    return !doc.definitions.some(def => isExecutableDefinitionNode(def));
}
function expandImports(source, options) {
    const lines = source.split(/\r\n|\r|\n/);
    let outputCode = options.importHelpers
        ? `
    var useUnique = require('webql-tools-webpack-loader-runtime').useUnique;
    var unique = useUnique();
  `
        : `
    ${uniqueCode}
  `;
    lines.some(line => {
        if (line[0] === '#' && line.slice(1).split(' ')[0] === 'import') {
            const importFile = line.slice(1).split(' ')[1];
            const parseDocument = `require(${importFile})`;
            const appendDef = `doc.definitions = doc.definitions.concat(unique(${parseDocument}.definitions));`;
            outputCode += appendDef + os.EOL;
        }
        return line.length !== 0 && line[0] !== '#';
    });
    return outputCode;
}
function graphqlLoader(source) {
    this.cacheable();
    const options = this.query || {};
    let doc = parseDocument(source);
    const optimizers = [];
    if (options.noDescription) {
        optimizers.push(removeDescriptions);
    }
    if (options.noEmptyNodes) {
        optimizers.push(removeEmptyNodes);
    }
    if (optimizers.length > 0 && isSDL(doc)) {
        doc = optimizeDocumentNode(doc, optimizers);
    }
    const headerCode = `
    var doc = ${JSON.stringify(doc)};
  `;
    let outputCode = '';
    // Allow multiple query/mutation definitions in a file. This parses out dependencies
    // at compile time, and then uses those at load time to create minimal query documents
    // We cannot do the latter at compile time due to how the #import code works.
    const operationCount = doc.definitions.reduce((accum, op) => {
        if (op.kind === Kind.OPERATION_DEFINITION) {
            return accum + 1;
        }
        return accum;
    }, 0);
    function exportDefaultStatement(identifier) {
        if (options.esModule) {
            return `export default ${identifier}`;
        }
        return `module.exports = ${identifier}`;
    }
    if (operationCount > 1) {
        throw new Error('GraphQL Webpack Loader allows only for one GraphQL Operation per file');
    }
    outputCode += `
    ${exportDefaultStatement('doc')}
  `;
    const importOutputCode = expandImports(source, options);
    const allCode = [headerCode, importOutputCode, outputCode, ''].join(os.EOL);
    return allCode;
}

export default graphqlLoader;
//# sourceMappingURL=index.esm.js.map
